# =============================================================================
# Ural Various Utilities
# =============================================================================
#
# Miscellaneous utilities used throughout the library's code.
#
import re

from ural.patterns import PROTOCOL_RE

# PY2/PY3 compatible string_type...
string_type = str

try:
    string_type = basestring
except NameError:
    pass

# PY2/PY3 compatible urlparse
try:
    from urllib.parse import (
        parse_qs,
        parse_qsl,
        quote,
        unquote,
        urljoin,
        urlsplit,
        urlunsplit,
        SplitResult,
    )
except ImportError:
    from urllib import unquote, quote as original_quote

    ALWAYS_SAFE = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_.-~"

    def quote(string, safe="/"):
        if isinstance(string, unicode):
            safe_set = set(ord(c) for c in ALWAYS_SAFE)

            for c in safe:
                safe_set.add(ord(c))

            chars = [
                c if ord(c) in safe_set else "%{:02X}".format(ord(c)) for c in string
            ]

            return "".join(chars)

        return original_quote(string, safe)

    from urlparse import parse_qs, parse_qsl, urljoin, urlsplit, urlunsplit, SplitResult

MISTAKES_RE = re.compile(r"&amp(?:%3B|;)", re.I)

unshadowed_quote = quote


def safe_urlsplit(url, scheme="http"):
    if isinstance(url, SplitResult):
        return url

    if not re.match(PROTOCOL_RE, url):
        url = scheme + "://" + url

    splitted = urlsplit(url)

    return splitted


def pathsplit(urlpath):
    urlpath = urlpath.strip()

    if not urlpath or urlpath == "/":
        return []

    urlpath = urlpath.strip("/")

    return urlpath.split("/")


def urlpathsplit(url):
    parsed = safe_urlsplit(url)
    return pathsplit(parsed.path)


SLASH_SQUEEZE_RE = re.compile(r"\/{2,}")


def normpath(urlpath, drop_consecutive_slashes=True):
    if drop_consecutive_slashes:
        urlpath = SLASH_SQUEEZE_RE.sub("/", urlpath)

    segments = urlpath.split("/")
    segments = [segment + "/" for segment in segments[:-1]] + [segments[-1]]
    resolved = []

    for segment in segments:
        if segment in ("../", ".."):
            if resolved[1:]:
                resolved.pop()
        elif segment not in ("./", "."):
            resolved.append(segment)

    return "".join(resolved).rstrip("/")


def attempt_to_decode_idna(string):
    try:
        return string.encode("utf8").decode("idna")
    except UnicodeError:
        return string


def decode_punycode_hostname(hostname, as_parts=False):
    parts = []

    for part in hostname.split("."):
        puny_header = part[:4].lower()

        if puny_header == "xn--":
            part = attempt_to_decode_idna(puny_header + part[4:])

        parts.append(part)

    if as_parts:
        return parts

    return ".".join(parts)


def fix_common_query_mistakes(query):
    return re.sub(MISTAKES_RE, "&", query)


def safe_parse_qs(query):
    return parse_qs(fix_common_query_mistakes(query))


def add_query_argument(url, name, value=None, quote=True):
    if quote:
        name = unshadowed_quote(name)

    if value == True or value is None:
        arg = name
    else:
        if quote:
            value = unshadowed_quote(str(value))

        arg = name + "=" + value

    query = None
    fragment = None

    s = url.rsplit("#", 1)

    if len(s) > 1:
        url, fragment = s

    s = url.rsplit("?", 1)

    if len(s) > 1:
        url, query = s

    if query:
        query += "&" + arg
    else:
        query = arg

    url += "?" + query

    if fragment is not None:
        url += "#" + fragment

    return url



# =============================================================================
# Ural LRUTrie Classes
# =============================================================================
#
# Bunch of classes relying on LRUs and prefix trees to perform complex URL
# matching routines.
#
from functools import partial

from ural.classes import TrieDict
from ural.utils import string_type
from ural.lru.stems import normalized_lru_stems, lru_stems
from ural.lru.serialization import unserialize_lru


def ensure_lru_stems(lru):
    if isinstance(lru, string_type):
        return unserialize_lru(lru)

    return lru


def clean_trailing_path(stems):
    return [stem for stem in stems if stem != "p:"]


class LRUTrie(object):
    def __init__(self, suffix_aware=False):
        self._trie = TrieDict()
        self.lru_tokenizer = partial(lru_stems, suffix_aware=suffix_aware)

    def __len__(self):
        return len(self._trie)

    def set(self, url, metadata):
        stems = self.lru_tokenizer(url)
        stems = clean_trailing_path(stems)
        self._trie[stems] = metadata

    def __setitem__(self, url, metadata):
        return self.set(url, metadata)

    def match(self, url):
        stems = self.lru_tokenizer(url)
        stems = clean_trailing_path(stems)
        return self._trie.longest_matching_prefix_value(stems)

    def set_lru(self, lru, metadata):
        stems = ensure_lru_stems(lru)
        stems = clean_trailing_path(stems)
        self._trie[stems] = metadata

    def match_lru(self, lru):
        stems = ensure_lru_stems(lru)
        stems = clean_trailing_path(stems)
        return self._trie.longest_matching_prefix_value(stems)

    def __iter__(self):
        return self._trie.values()


class NormalizedLRUTrie(LRUTrie):
    def __init__(self, suffix_aware=False, **kwargs):
        self._trie = TrieDict()
        self.lru_tokenizer = partial(
            normalized_lru_stems, suffix_aware=suffix_aware, **kwargs
        )



# =============================================================================
# Ural LRU Serialization Functions
# =============================================================================
#
# Functions related to LRU serialization.
#
import re

SERIALIZED_LRU_SPLITTER_RE = re.compile(r"\|(?=[shtpqfuw]:)")


def serialize_lru(stems):
    return "|".join(stems) + "|"


def unserialize_lru(lru):
    lru = lru.rstrip("|")
    return SERIALIZED_LRU_SPLITTER_RE.split(lru)



# -*- coding: utf-8 -*-
# =============================================================================
# Ural Format Url Function
# =============================================================================
#
# Function used to format urls conveniently.
#
from __future__ import unicode_literals

try:
    from collections.abc import Iterable
except ImportError:
    Iterable = (tuple, list, set, frozenset)

from ural.utils import string_type, quote


def format_query_argument(key, value, format_value=None):
    if format_value is not None:
        value = format_value(key, value)

    if value is True:
        return quote(key)

    return quote(key) + "=" + quote(str(value))


def format_url(
    base_url, path=None, args=None, format_arg_value=None, fragment=None, ext=None
):
    url = base_url

    # Path
    if isinstance(path, string_type):
        url = base_url.rstrip("/") + "/" + path.lstrip("/")
    elif isinstance(path, Iterable):
        url = base_url.rstrip("/") + "/" + "/".join(path).lstrip("/")
    elif path is not None:
        raise TypeError("path should be a string or an iterable of path items")

    # Extension
    # NOTE: this can cause issues in some configurations, beware
    if ext is not None:
        url += "." + ext.lstrip(".")

    # Arguments
    if args is not None:
        items = sorted(
            (
                format_query_argument(k, v, format_arg_value)
                for k, v in args.items()
                if v is not None and v is not False
            )
        )

        if items:
            url += "?" + ("&".join(items))

    # Fragment
    if fragment is not None:
        url += "#" + fragment.lstrip("#")

    return url


class URLFormatter(object):
    BASE_URL = None

    def __init__(
        self,
        base_url=None,
        path=None,
        args=None,
        format_arg_value=None,
        fragment=None,
        ext=None,
    ):
        self.base_url = self.BASE_URL if base_url is None else base_url
        self.path = path
        self.args = args
        self._format_arg_value = format_arg_value
        self.fragment = fragment
        self.ext = ext

    def format_arg_value(self, k, v):
        return v

    def format(
        self,
        base_url=None,
        path=None,
        args=None,
        format_arg_value=None,
        fragment=None,
        ext=None,
    ):
        base_url = self.base_url if base_url is None else base_url
        path = self.path if path is None else path

        if format_arg_value is None:
            format_arg_value = self._format_arg_value

        if format_arg_value is None:
            format_arg_value = self.format_arg_value

        fragment = self.fragment if fragment is None else fragment

        # Args are merged
        if args is None:
            args = self.args
        else:
            if self.args is not None:
                new_args = self.args.copy()
                new_args.update(args)
                args = new_args

        return format_url(
            base_url=base_url,
            path=path,
            args=args,
            format_arg_value=format_arg_value,
            fragment=fragment,
            ext=ext,
        )

    def __call__(
        self,
        base_url=None,
        path=None,
        args=None,
        format_arg_value=None,
        fragment=None,
        ext=None,
    ):
        return self.format(
            base_url=base_url,
            path=path,
            args=args,
            format_arg_value=format_arg_value,
            fragment=fragment,
            ext=ext,
        )



# =============================================================================
# Ural Instagram-related heuristic functions
# =============================================================================
#
# Collection of functions related to Instagram urls.
#
import re
from collections import namedtuple

from ural.utils import pathsplit, safe_urlsplit, SplitResult
from ural.patterns import DOMAIN_TEMPLATE

INSTAGRAM_POST_SHORTCODE_RE = re.compile(r"^[a-zA-Z0-9_\-]+$")
INSTAGRAM_USERNAME_RE = re.compile(r"^[a-zA-Z0-9_\-\.]+$")
INSTAGRAM_DOMAIN_RE = re.compile(r"instagram.com$", re.I)
INSTAGRAM_URL_RE = re.compile(DOMAIN_TEMPLATE % r"(?:[^.]+\.)*instagram.com", re.I)
INSTAGRAM_NOT_A_USER_SET = {
    "accounts",
    "ads",
    "direct",
    "directory",
    "emails",
    "explore",
    "push",
    "reels",
    "session",
    "settings",
    "stories",
    "web",
    "your_activity",
}

InstagramUser = namedtuple("InstagramUser", ["name"])
InstagramPost = namedtuple("InstagramPost", ["id", "name"])
InstagramReel = namedtuple("InstagramReel", ["id"])


def is_instagram_post_shortcode(value):
    return bool(re.search(INSTAGRAM_POST_SHORTCODE_RE, value))


def is_instagram_username(value):
    return (
        bool(re.search(INSTAGRAM_USERNAME_RE, value))
        and value not in INSTAGRAM_NOT_A_USER_SET
    )


def is_instagram_url(url):
    """
    Function returning whether the given url is a valid Instagram url.

    Args:
        url (str): Url to test.

    Returns:
        bool: Whether given url is from Instagram.

    """
    if isinstance(url, SplitResult):
        return bool(re.search(INSTAGRAM_DOMAIN_RE, url.hostname))

    return bool(re.match(INSTAGRAM_URL_RE, url))


def parse_instagram_url(url):
    """
    Function parsing the given url and returning either a InstagramPost,
    InstagramUser or None if nothing of information could be found.

    Args:
        url (str): Url to parse.

    """
    if not is_instagram_url(url):
        return None

    parsed = safe_urlsplit(url)
    path = pathsplit(parsed.path)

    if not path:
        return None

    if path[0] == "p" and len(path) >= 2:

        if is_instagram_post_shortcode(path[1]):
            return InstagramPost(id=path[1], name=None)

        return None

    elif path[0] == "reel" and len(path) >= 2:

        if is_instagram_post_shortcode(path[1]):
            return InstagramReel(id=path[1])

        return None

    elif path[0] == "reels" and len(path) >= 2:

        if is_instagram_post_shortcode(path[1]) and path[1] != "videos":
            return InstagramReel(id=path[1])

        elif (
            len(path) >= 3
            and path[1] == "videos"
            and is_instagram_post_shortcode(path[2])
            and path[2] != "videos"
        ):
            return InstagramReel(id=path[2])

        return None

    elif is_instagram_username(path[0]):

        if len(path) >= 3 and path[1] == "p" and is_instagram_post_shortcode(path[2]):
            return InstagramPost(id=path[2], name=path[0])

        return InstagramUser(name=path[0])

    return None


def extract_username_from_instagram_url(url):
    parsed = parse_instagram_url(url)

    if not isinstance(parsed, (InstagramPost, InstagramUser)):
        return

    return parsed.name



# -*- coding: utf-8 -*-
# =============================================================================
# Ural Shortened Url Testing Function
# =============================================================================
#
# Function testing whether the given url is probably a shortened url or not.
#
import re

from ural.classes.hostname_trie_set import HostnameTrieSet
from ural.is_homepage import is_homepage
from ural.utils import safe_urlsplit

DOMAIN_STARTS_L_RE = re.compile(r"^/[0-9a-zA-Z]{3,}/?$")

# Some domains taken from: https://github.com/timleland/url-shorteners
SHORTENER_DOMAINS = [
    "0rz.tw",
    "1-url.net",
    "1001.tl",
    "126.am",
    "1link.in",
    "1tk.us",
    "1un.fr",
    "1url.com",
    "1url.cz",
    "1wb2.net",
    "2.gp",
    "2.ht",
    "20m.es",
    "23o.net",
    "2ad.in",
    "2big.at",
    "2doc.net",
    "2fear.com",
    "2kgam.es",
    "2long.cc",
    "2no.co",
    "2pl.us",
    "2tu.us",
    "2ty.in",
    "2u.xf.cz",
    "2ya.com",
    "3.ly",
    "307.to",
    "3ds.com",
    "3ra.be",
    "3x.si",
    "4i.ae",
    "4ms.me",
    "4sq.com",
    "4url.cc",
    "4view.me",
    "53eig.ht",
    "5em.cz",
    "5url.net",
    "5z8.info",
    "6fr.ru",
    "6g6.eu",
    "6url.com",
    "7.ly",
    "76.gd",
    "77.ai",
    "7fth.cc",
    "7li.in",
    "7ny.tv",
    "7vd.cn",
    "8u.cz",
    "944.la",
    "98.to",
    "9m.no",
    "9qr.de",
    "a.co",
    "a.gg",
    "a.msn.com",
    "a.nf",
    "a.sgdf.fr",
    "a0.fr",
    "a2a.me",
    "a42.fr",
    "aa.cx",
    "ab.co",
    "abbr.sk",
    "abbrr.com",
    "abc7.la",
    "abcn.ws",
    "abcurl.net",
    "abe.ma",
    "abnb.me",
    "aca.re",
    "acani.me",
    "acmi.land",
    "acortar.link",
    "act.gp",
    "acver.fr",
    "ad-med.cz",
    "ad.vu",
    "ad5.eu",
    "ad7.biz",
    "adb.ug",
    "adec.co",
    "adf.ly",
    "adfa.st",
    "adfly.fr",
    "adfoc.us",
    "adjix.com",
    "adli.pw",
    "admy.link",
    "adn.fm",
    "adv.li",
    "afx.cc",
    "ajb.me",
    "aje.io",
    "ajn.me",
    "aka.gr",
    "aka.ms",
    "al.ly",
    "album.ink",
    "alil.in",
    "altav.eu",
    "altereco.media",
    "alturl.com",
    "amn.st",
    "amp.gs",
    "amz.run",
    "amzn.asia",
    "amzn.eu",
    "amzn.to",
    "any.gs",
    "apne.ws",
    "apo-opa.info",
    "app.link",
    "apple.co",
    "aqva.pl",
    "ar.gy",
    "arab.news",
    "arcg.is",
    "ares.tl",
    "arst.ch",
    "assnat.fr",
    "asso.in",
    "atmlb.com",
    "atres.red",
    "atu.ca",
    "au.ms",
    "audiof.com",
    "audmns.com",
    "ava.st",
    "awr.as",
    "ayt.fr",
    "azali.fr",
    "azc.cc",
    "b00.fr",
    "b23.ru",
    "b23.tv",
    "b2l.me",
    "b54.in",
    "bacn.me",
    "baid.us",
    "barca.link",
    "bars.tl",
    "bbc.in",
    "bc.vc",
    "bcool.bz",
    "bddy.me",
    "bee4.biz",
    "beth.games",
    "bigaston.link",
    "bim.im",
    "binged.it",
    "bit.do",
    "bit.ly",
    "bitchute.xyz",
    "bitly.com",
    "bitly.ws",
    "bitw.in",
    "bityl.co",
    "bizj.us",
    "bkite.com",
    "blap.net",
    "blbrd.cm",
    "ble.pl",
    "blgs.co",
    "blip.tv",
    "blizz.ly",
    "bloat.me",
    "bmsp.in",
    "bnc.lt",
    "bnent.eu",
    "bnpp.lk",
    "boi.re",
    "bote.me",
    "bougn.at",
    "boxd.it",
    "br4.in",
    "bravo.ly",
    "bre.is",
    "brk.to",
    "brzu.net",
    "bsa.ly",
    "budurl.com",
    "buff.ly",
    "buk.me",
    "bul.lu",
    "burnurl.com",
    "bxl.me",
    "bzh.me",
    "c-o.in",
    "c.bnf.fr",
    "c11.kr",
    "c3po.link",
    "cachor.ro",
    "cam4.fr",
    "can.al",
    "canurl.com",
    "capfr.link",
    "captur.in",
    "catchylink.com",
    "cbs.so",
    "cbsn.ws",
    "cbug.cc",
    "cc.cc",
    "cciparisidf.fr",
    "ccj.im",
    "ccompt.es",
    "cd.leg.br",
    "cf.ly",
    "cf2.me",
    "cf6.co",
    "chck.me",
    "chilp.it",
    "chl.li,",
    "chng.it",
    "chocobp.com",
    "chrk.tv",
    "chzb.gr",
    "cirk.me",
    "cjb.net",
    "cl.lk",
    "cl.ly",
    "clar.in",
    "clck.ru",
    "cldp.ly",
    "cleanuri.com",
    "cli.gs",
    "cliccami.info",
    "clickmeter.com",
    "clickmetertracking.com",
    "clickthru.ca",
    "clikk.in",
    "clop.in",
    "cmpdnt.cc",
    "cn86.org",
    "cna.asia",
    "cnb.cx",
    "cnet.co",
    "cnhv.co",
    "cnn.it",
    "coinurl.com",
    "colibris.link",
    "conta.cc",
    "cor.sr",
    "cort.as",
    "cos.lv",
    "cot.ag",
    "couic.fr",
    "cpu.pm",
    "cr.tl",
    "crdrv.co",
    "crks.me",
    "crwd.fr",
    "cstu.io",
    "ctt.ec",
    "ctvr.us",
    "cudder.it",
    "cur.lv",
    "curl.im",
    "cut.ci",
    "cut.pe",
    "cut.sk",
    "cutt.eu",
    "cutt.ly",
    "cutt.us",
    "cutu.me",
    "cuturl.com",
    "cybr.fr",
    "cyonix.to",
    "d75.eu",
    "daa.pl",
    "dai.ly",
    "dailysign.al",
    "dainik-b.in",
    "dart.ws",
    "db.tt",
    "dbd.game",
    "dd.ma",
    "ddp.net",
    "ddsevilla.info",
    "decenturl.com",
    "del.ly",
    "dell.to",
    "dfl8.me",
    "dft.ba",
    "dgl.li",
    "di.sn",
    "digbig.com",
    "digg.com",
    "digimind.co",
    "digt.al",
    "disq.us",
    "dlbs.fr",
    "dld.bz",
    "dlvr.it",
    "dncgastrnt.co",
    "do.my",
    "docdro.id",
    "doiop.com",
    "dolp.cc",
    "doopin.net",
    "dopen.us",
    "dopice.sk",
    "dozz.es",
    "droid.ws",
    "dsmrq.es",
    "dv.gd",
    "dvelo.me",
    "dwarfurl.com",
    "dy.fi",
    "dy.si",
    "dyo.gs",
    "e.vg",
    "e37.eu",
    "earthquak.es",
    "easyuri.com",
    "easyurl.com",
    "easyurl.net",
    "ebx.sh",
    "econ.st",
    "ecra.se",
    "ed.gr",
    "edcollector.fr",
    "eef.li",
    "eepurl.com",
    "efs.link",
    "egtks.com",
    "eiie.io",
    "eldbt.com",
    "ellemag.co",
    "elpais.uy",
    "elperiodi.co",
    "ely.re",
    "emojilink.me",
    "energy-citi.es",
    "eng.mg",
    "engt.co",
    "eonli.ne",
    "ept.ms",
    "eptwitter.eu",
    "equid.la",
    "erax.cz",
    "erw.cz",
    "es.pn",
    "eset.la",
    "esyurl.com",
    "etcse.fr",
    "etlink.fr",
    "etoro.tw",
    "etsy.me",
    "europa.eu",
    "eweri.com",
    "ewerl.com",
    "ex9.co",
    "ezurl.cc",
    "f1.com",
    "f24.my",
    "fa.b",
    "fa.by",
    "fal.cn",
    "fanpa.ge",
    "fav.me",
    "fb.me",
    "fb.watch",
    "fbshare.me",
    "fcld.ly",
    "fdip.fr",
    "feedproxy.google.com",
    "ff.im",
    "fff.re",
    "fff.to",
    "fff.wf",
    "fhurl.com",
    "filz.fr",
    "fire.to",
    "firsturl.de",
    "firsturl.net",
    "flashsco.re",
    "flic.kr",
    "flip.it",
    "fliz.ly",
    "flq.us",
    "fly2.ws",
    "fnk.es",
    "foe.hn",
    "folu.me",
    "fon.gs",
    "fr.rog.gg",
    "fr24.com",
    "frama.link",
    "freak.to",
    "freze.it",
    "from.ubs",
    "fur.ly",
    "fuseurl.com",
    "fuzzy.to",
    "fw.to",
    "fwd4.me",
    "fwib.net",
    "fxn.ws",
    "g.co",
    "g.pe",
    "g.ro.lt",
    "g00.me",
    "g2g.to",
    "gal.xyz",
    "gcn.eu",
    "gd.lu",
    "geni.us",
    "gerd.fm",
    "gf.me",
    "gg.gg",
    "git.io",
    "giv.gg",
    "gizmo.do",
    "gl.am",
    "glo.bo",
    "gloss.bz",
    "glplive.org",
    "gma.abc",
    "go.9nl.com",
    "go.asse.fr",
    "go.cegeka.com",
    "go.ign.com",
    "go.letelegramme.fr",
    "go.nasa.gov",
    "go.one.org",
    "go.shr.lc",
    "go.tass.ru",
    "go.usa.gov",
    "go2.me",
    "go2cut.com",
    "gofund.me",
    "goha.bs",
    "goo-gl.me",
    "goo.gl",
    "goo.gle",
    "goo.lu",
    "good.ly",
    "goodsmile.link",
    "goog.le",
    "goshrink.com",
    "got.cr",
    "gotrim.me",
    "gowat.ch",
    "gplnk.com",
    "gq.mn",
    "grabify.link",
    "grem.io",
    "gri.ms",
    "guiama.is",
    "gurl.es",
    "h-c.info",
    "hadej.co",
    "han.gl",
    "hbsd.shop",
    "hc.hockey",
    "hec.su",
    "hellotxt.com",
    "heral.do",
    "hex.io",
    "hide.my",
    "hiderefer.com",
    "hill.cm",
    "hjkl.fr",
    "hmm.ph",
    "hmstr.biz",
    "hndcp.fr",
    "hollywood.li",
    "hops.me",
    "hover.com",
    "href.in",
    "href.li",
    "hrvd.me",
    "hsblinks.com",
    "hsnob.co",
    "ht.ly",
    "htl.li",
    "httpslink.com",
    "htxt.it",
    "hubs.la",
    "hubs.li",
    "hubs.ly",
    "huff.to",
    "hugeurl.com",
    "hulu.com",
    "hurl.it",
    "hurl.me",
    "hurl.ws",
    "hyperurl.co",
    "i-2.co",
    "i99.cz",
    "iaddict.co",
    "ibit.ly",
    "ibm.co",
    "icanhaz.com",
    "icit.fr",
    "ick.li",
    "icks.ro",
    "idek.net",
    "iea.li",
    "ift.tt",
    "ihe.art",
    "ihr.fm",
    "iiiii.in",
    "iky.fr",
    "ilix.in",
    "imdb.to",
    "indeedhi.re",
    "info.ms",
    "infob.ae",
    "infomi.gr",
    "ino.to",
    "inq.news",
    "inreply.to",
    "instagr.am",
    "interc.pt",
    "intg.ink",
    "invst.ly",
    "io.webhelp.com",
    "is.gd",
    "iscool.net",
    "isra.li",
    "isra.liiterasi.net",
    "isw.pub",
    "iterasi.net",
    "itm.im",
    "its.my",
    "ity.im",
    "ix.lt",
    "ix.sk",
    "j.gs",
    "j.mp",
    "ja.ma",
    "jb.gg",
    "jdem.cz",
    "jieb.be",
    "jijr.com",
    "jmp2.net",
    "jp22.net",
    "jpeg.ly",
    "jqw.de",
    "just.as",
    "juve.it",
    "jwatch.us",
    "kas.pr",
    "kask.us",
    "kbn.im",
    "kd2.org",
    "kfd.pl",
    "kgeek.co",
    "kissa.be",
    "kl.am",
    "klck.me",
    "kloo.me",
    "korta.nu",
    "kr3w.de",
    "krat.si",
    "kratsi.cz",
    "krod.cz",
    "krunchd.com",
    "kuc.cz",
    "kutt.it",
    "kxb.me",
    "l-fff.co",
    "l-k.be",
    "l.gg",
    "l.linklyhq.com",
    "l8r.it",
    "l9.fr",
    "l9k.net",
    "la4.cl",
    "labur.eus",
    "lat.ms",
    "lavoz.gal",
    "lawl.es",
    "lbrs.ca",
    "lc-s.co",
    "lc.cx",
    "lcut.in",
    "letop10.",
    "lety.io",
    "lew.ag",
    "lext.so",
    "lgbtq.to",
    "libero.it",
    "lick.my",
    "lien.li",
    "lien.pl",
    "lienfy.com",
    "lifehac.kr",
    "liip.to",
    "liltext.com",
    "lin.cr",
    "lin.ee",
    "lin.io",
    "link.emakinainflux.com",
    "link.infini.fr",
    "link.medium.com",
    "linkbee.com",
    "linkbun.ch",
    "linkn.co",
    "linkto.cm",
    "listani.me",
    "liurl.cn",
    "llu.ch",
    "ln-s.net",
    "ln-s.ru",
    "lnk.bvh.fyi",
    "lnk.co",
    "lnk.gd",
    "lnk.in",
    "lnk.ly",
    "lnk.ms",
    "lnk.sk",
    "lnkd.in",
    "lnked.in",
    "lnki.nl",
    "lnkiy.com,",
    "lnkr.fm",
    "lnks.fr",
    "lnkurl.com",
    "lnky.fr",
    "lnky.jp",
    "lnp.sn",
    "lnr.app",
    "loom.ly",
    "loopt.us",
    "lp.ca",
    "lp25.fr",
    "lru.jp",
    "lrzn.es",
    "lsdl.es",
    "lstu.fr",
    "lsul.su",
    "lt.tl",
    "lttr.ai",
    "lurl.no",
    "lvvk.com",
    "lynk.my",
    "lyonna.is",
    "m-et-d.link",
    "m1p.fr",
    "m3mi.com",
    "macte.ch",
    "maddy.news",
    "mailchi.mp",
    "make.my",
    "mash.to",
    "maza.rs",
    "mb4.me",
    "mbe.io",
    "mcaf.ee",
    "mck.co",
    "mdl29.net",
    "mdn.tv",
    "mdpt.fr",
    "merky.de",
    "metamark.net",
    "mic.fr",
    "michel.in",
    "micro.mn",
    "migre.me",
    "mile.io",
    "minilien.com",
    "miniurl.be",
    "miniurl.com",
    "minu.me",
    "minurl.fr",
    "mke.me",
    "mla.bs",
    "mmt.fans",
    "moby.to",
    "mol.im",
    "mon.actu.io",
    "monurl.ca",
    "moourl.com",
    "more.sh",
    "mrte.ch",
    "msft.it",
    "msft.social",
    "msi.gm",
    "mtr.cool",
    "mut.lu",
    "my.sociabble.com",
    "my.wolters-kluwer.fr",
    "myjob.scot",
    "myloc.me",
    "myumi.ch",
    "myurl.in",
    "n.pr",
    "n9.cl",
    "nac.ca",
    "nar.city",
    "natgeo.com",
    "naver.me",
    "nbc.co",
    "nbcnews.to",
    "nblo.gs",
    "ne1.net",
    "net.ms",
    "net46.net",
    "nico.ms",
    "nicou.ch",
    "nie.mn",
    "nig.gr",
    "ninten.do",
    "niny.io",
    "njx.me",
    "nn.nf",
    "non.li",
    "not.my",
    "notlong.com",
    "nov.io",
    "nowth.is",
    "nq.st",
    "nsfw.in",
    "ntdo.be",
    "ntdo.fr",
    "nutshellurl.com",
    "nvda.ws",
    "nvth.fr",
    "nxy.in",
    "nyer.cm",
    "nym.ag",
    "nyti.ms",
    "o-x.fr",
    "oc1.us",
    "ocul.us",
    "oe.cd",
    "ogcn.fr",
    "okok.fr",
    "olj.me",
    "olplus.fr",
    "om.ly",
    "omf.gd",
    "omoikane.net",
    "on.cnn.com",
    "on.ft.com",
    "on.mktw.net",
    "on.msnbc.com",
    "on.unesco.org",
    "onelink.me",
    "onforb.es",
    "options-po.li",
    "oran.ge",
    "orlo.uk",
    "orz.se",
    "ottsens.com",
    "ou.af",
    "ou.gd",
    "oua.be",
    "ouo.io",
    "ouvaton.link",
    "ow.ly",
    "owl.li",
    "p.pw",
    "p3t.co",
    "p4k.in",
    "para.pt",
    "parky.tv",
    "past.is",
    "paypal.me",
    "pbi.bz",
    "pd.am",
    "pdca.st",
    "pdh.co",
    "pens.pe",
    "peta.vg",
    "petafr.vg",
    "ph.dog",
    "ph.ly",
    "pic.gd",
    "pich.in",
    "picta.fr",
    "pin.it",
    "pin.st",
    "ping.fm",
    "piurl.com",
    "pkmn.news",
    "play.st",
    "playst.cc",
    "pli.gs",
    "plots.fr",
    "plstrw.rs",
    "pnt.me",
    "po.st",
    "pod.fo",
    "politi.co",
    "poprl.com",
    "pos.li",
    "post.ly",
    "posted.at",
    "posty.pe",
    "pp.gg",
    "ppfr.it",
    "ppst.me",
    "ppt.cc",
    "ppt.li",
    "prejit.cz",
    "prez.ly",
    "prf.hn",
    "prn.to",
    "profile.to",
    "propub.li",
    "psta.rs",
    "ptab.it",
    "ptiturl.com",
    "ptm.ro",
    "pub.vitrue.com",
    "pubg.info",
    "pulse.ly",
    "pw2.ro",
    "pwc.to",
    "pwne.ws",
    "py6.ru",
    "q.gs",
    "qbn.ru",
    "qdlink.co",
    "qicute.com",
    "qlnk.net",
    "qob.uz",
    "qqc.co",
    "qr.net",
    "qrco.de",
    "qrtag.fr",
    "qte.me",
    "qu.tc",
    "quip-art.com",
    "qxp.cz",
    "qxp.sk",
    "qy.fi",
    "r.im",
    "rb.gy",
    "rb6.co",
    "rb6.me",
    "rc.ca",
    "rcknr.io",
    "rdo.to",
    "rdz.me",
    "read.bi",
    "readerlinks.com",
    "readthis.ca",
    "reallytinyurl.com",
    "rebelne.ws",
    "rebrand.ly",
    "rebrandlydomain.com",
    "redir.ec",
    "redir.fr",
    "redirects.ca",
    "redirx.com",
    "redu.it",
    "ref.so",
    "reise.lc",
    "rel.ink",
    "relink.fr",
    "reso.ly",
    "retwt.me",
    "reut.rs",
    "rfi.my",
    "rflsnk.link",
    "rfr.bz",
    "ri.ms",
    "rickroll.it",
    "riot.com",
    "riz.cz",
    "riz.gd",
    "rl.gg",
    "rng.rs",
    "roc.gg",
    "rod.gs",
    "roflc.at",
    "rog.gg",
    "rs-thales.com",
    "rsmonkey.com",
    "rt.nu",
    "rt.se",
    "rt.tc",
    "rtrv.in",
    "ru.ly",
    "rubyurl.com",
    "rurl.org",
    "rww.tw",
    "s-url.fr",
    "s.42l.fr",
    "s.bulk.ly",
    "s.click.aliexpress.com",
    "s.id",
    "s.swissin.fo",
    "s.team",
    "s4c.in",
    "s7y.us",
    "sabqp.co",
    "safe.mn",
    "sagyap.tk",
    "sameurl.com",
    "sbee.link",
    "sc.mp",
    "schrts.co",
    "sci3nc.es",
    "scim.ag",
    "sco.lt",
    "scq.io",
    "scuf.co",
    "sdu.sk",
    "sdut.us",
    "seeme.at",
    "segidio.it",
    "segue.se",
    "sg.fr",
    "sh.st",
    "shar.as",
    "shar.es",
    "sharee.me",
    "sharein.com",
    "sharetabs.com",
    "shdw.me",
    "shink.de",
    "shope.ee",
    "shorl.com",
    "short.cc",
    "short.ie",
    "short.nr",
    "short.pk",
    "short.to",
    "shorte.st",
    "shortlink.store",
    "shortlinks.co.uk",
    "shortna.me",
    "shorturl.at",
    "shorturl.com",
    "shoturl.us",
    "shout.to",
    "show.my",
    "shrinkee.com",
    "shrinkify.com",
    "shrinkr.com",
    "shrinkster.com",
    "shrinkurl.in",
    "shrt.fr",
    "shrt.in",
    "shrt.st",
    "shrtco.de",
    "shrten.com",
    "shrunkin.com",
    "shw.me",
    "shy.si",
    "sicax.net",
    "simurl.com",
    "sina.lt",
    "sk.gy",
    "sk.mu",
    "skfb.ly",
    "sking.fr",
    "skr.sk",
    "skroc.pl",
    "slate.me",
    "smallr.com",
    "smarturl.it",
    "smll.co",
    "smpl.is",
    "smsh.me",
    "smurl.name",
    "sn.cf",
    "sn.im",
    "sn.vc",
    "snipr.com",
    "snipurl.com",
    "snkrne.ws",
    "snsw.us",
    "snurl.com",
    "so.arte",
    "soc.cr",
    "soc.fm",
    "social.ora.cl",
    "socy.es",
    "sohub.io",
    "soo.gd",
    "sovrn.co",
    "sp2.ro",
    "spacekaren.sucks",
    "spedr.com",
    "spkl.io",
    "spn.sr",
    "spoti.fi",
    "spr.ly",
    "sptfy.com",
    "sq6.ru",
    "sqex.link",
    "sqex.to",
    "sqrl.it",
    "srnk.net",
    "srs.li",
    "ssd.fr",
    "ssl.gs",
    "starturl.com",
    "strms.net",
    "sturly.com",
    "su.pr",
    "sumof.us",
    "supr.cl",
    "sur.om.fr",
    "surl.co.uk",
    "surl.hu",
    "surl.me",
    "sux.cz",
    "svlint.org",
    "swki.me",
    "swll.to",
    "sy.pe",
    "sz.de",
    "t.cn",
    "t.co",
    "t.lh.com",
    "t.ly",
    "t.me",
    "t1p.de",
    "t2m.io",
    "ta.gd",
    "tabzi.com",
    "tau.pe",
    "tbd.ly",
    "tblink.co",
    "tcrn.ch",
    "tdjt.cz",
    "tek.io",
    "tex.nz",
    "tgr.me",
    "tgr.ph",
    "theatln.tc",
    "theins.ru",
    "thesa.us",
    "thesco.re",
    "thinfi.com",
    "thls.co",
    "thr.cm",
    "thrdl.es",
    "ti.me",
    "tidd.ly",
    "tighturl.com",
    "tin.li",
    "tini.cc",
    "tiniuri.com",
    "tiny.cc",
    "tiny.lt",
    "tiny.ly",
    "tiny.ms",
    "tiny.one",
    "tiny.pl",
    "tiny123.com",
    "tinyarro.ws",
    "tinyl.io",
    "tinylf.com",
    "tinylink.in",
    "tinytw.it",
    "tinyuri.ca",
    "tinyurl.com",
    "tinyurl.hu",
    "tinyvid.io",
    "tixsu.com",
    "tk.",
    "tl.gd",
    "tldr.sk",
    "tldrify.com",
    "tllg.net",
    "tlmdo.co",
    "tmblr.co",
    "tmi.me",
    "tmz.me",
    "tnij.org",
    "tnvge.co",
    "tnw.to",
    "tny.com",
    "tny.cz",
    "tny.im",
    "to.ly",
    "to8.cc",
    "togoto.us",
    "tohle.de",
    "toi.in",
    "totalenergi.es",
    "totc.us",
    "toysr.us",
    "tpm.ly",
    "tpmr.com",
    "tr.ee",
    "tr.im",
    "tr.my",
    "tr5.in",
    "tra.kz",
    "traceurl.com",
    "trbr.io",
    "trck.me",
    "trib.al",
    "trick.ly",
    "trkr.ws",
    "trunc.it",
    "tsumeart.co",
    "tun.in",
    "tuppu.net",
    "turo.us",
    "tv3.video",
    "tv5m.tv",
    "tweetburner.com",
    "twet.fr",
    "twhub.com",
    "twi.im",
    "twib.in",
    "twirl.at",
    "twit.ac",
    "twitclicks.com",
    "twitterpan.com",
    "twitterurl.net",
    "twitterurl.org",
    "twitthis.com",
    "twiturl.de",
    "twlr.me",
    "twurl.cc",
    "twurl.nl",
    "u.afp.com",
    "u.mavrev.com",
    "u.nu",
    "u.to",
    "u2l.fr",
    "u6e.de",
    "u76.org",
    "ub0.cc",
    "ubb.link",
    "ubi.li",
    "ubxm.fr",
    "uby.es",
    "ucam.me",
    "ufcqc.link",
    "ug.cz",
    "ulmt.in",
    "ulu.lu",
    "unf.pa",
    "unh.cr",
    "uni.cf",
    "unlc.us",
    "untp.beer",
    "unwo.men",
    "updating.me",
    "upzat.com",
    "ur1.ca",
    "url.az",
    "url.co.uk",
    "url.ie",
    "url.inrae.fr",
    "url.kr",
    "url2.fr",
    "url360.me",
    "url4.eu",
    "url5.org",
    "urlao.com",
    "urlborg.com",
    "urlbrief.com",
    "urlcover.com",
    "urlcut.com",
    "urlenco.de",
    "urlhawk.com",
    "urli.nl",
    "urlin.it",
    "urlkiss.com",
    "urlkr.com",
    "urlot.com",
    "urlpire.com",
    "urlr.me",
    "urls.fr",
    "urls.im",
    "urlshorteningservicefortwitter.com",
    "urlx.ie",
    "urlx.org",
    "urlz.fr",
    "urlz.host",
    "urlzen.com",
    "urub.us",
    "usat.ly",
    "use.my",
    "uslsc.cr",
    "utfg.sk",
    "v.gd",
    "v.ht",
    "v5.gd",
    "vaaa.fr",
    "valv.im",
    "vaza.me",
    "vb.ly",
    "vbet.biz",
    "vbly.us",
    "vd55.com",
    "vdn.lv",
    "verd.in",
    "vgn.am",
    "vgn.me",
    "viid.me",
    "virl.com",
    "vl.am",
    "vm.lc",
    "vm.tiktok.com",
    "vni.fr",
    "vntyfr.com",
    "vogue.cm",
    "vov.li",
    "vsll.eu",
    "vt802.us",
    "vu.fr",
    "vult.re",
    "vur.me",
    "vv.vg",
    "w-j.co",
    "w.lpnt.fr",
    "w.tt",
    "w.wiki",
    "w1p.fr",
    "w3d.to",
    "w3t.org",
    "w55.de",
    "wa.link",
    "wa.me",
    "waa.ai",
    "wapo.st",
    "wapurl.co.uk",
    "wb1.eu",
    "wclink.co",
    "wdwnt.news",
    "we.tl",
    "web99.eu",
    "wed.li",
    "weez.li",
    "wef.ch",
    "weverse.onelink.me",
    "wh3.io",
    "whln.eu",
    "wideo.fr",
    "wipi.es",
    "wn.nr",
    "workfor.us",
    "wow.link",
    "wp.me",
    "wrfr.me",
    "write.as",
    "wrld.bg",
    "wt.link",
    "wtc.la",
    "wu.cz",
    "ww7.fr",
    "www.sharee.me",
    "www.tmz.me",
    "wwy.me",
    "x.co",
    "x.nu",
    "x.se",
    "x.vu",
    "x10.mx",
    "x2c.eu",
    "x2c.eumx",
    "xaddr.com",
    "xav.cc",
    "xbx.lv",
    "xeeurl.com",
    "xfru.it",
    "xgd.in",
    "xhtxs.cn",
    "xib.me",
    "xl8.eu",
    "xoe.cz",
    "xq.ms",
    "xr.com",
    "xrb.link",
    "xrl.in",
    "xrl.us",
    "xt3.me",
    "xua.me",
    "xub.me",
    "xurl.es",
    "xurl.jp",
    "xurls.co",
    "xzb.cc",
    "y.ahoo.it",
    "y2u.be",
    "yagoa.fr",
    "yagoa.me",
    "yatuc.com",
    "yau.sh",
    "ye.pe",
    "yeca.eu",
    "yect.com",
    "yep.it",
    "yfrog.com",
    "yhoo.it",
    "yiyd.com",
    "yogh.me",
    "yon.ir",
    "youfap.me",
    "youtu.be",
    "ysear.ch",
    "yuarel.com",
    "yweb.com",
    "yyv.co",
    "z0p.de",
    "z9.fr",
    "zapit.nu",
    "zcu.io",
    "zeek.ir",
    "zi.ma",
    "zi.mu",
    "zi.pe",
    "zip.net",
    "zipmyurl.com",
    "zkr.cz",
    "zkrat.me",
    "zkrt.cz",
    "zoodl.com",
    "zpag.es",
    "zpr.io",
    "zsms.net",
    "zti.me",
    "zubb.it",
    "zud.me",
    "zurl.io",
    "zurl.ws",
    "zws.im",
    "zxq.net",
    "zyva.org",
    "zz.gd",
    "zzang.kr",
    "zzb.bz",
    "›.ws",
    "✩.ws",
    "✿.ws",
    "❥.ws",
    "➔.ws",
    "➞.ws",
    "➡.ws",
    "➨.ws",
    "➯.ws",
    "➹.ws",
    "➽.ws",
    "🍊.ws",
    "🏹.to",
    "😻🍕.ws",
]

# NOTE: we use a trie to perform efficient queries and so we don't
# need to test every domain/subdomain linearly
SHORTENER_DOMAINS_TRIE = HostnameTrieSet()

for domain in SHORTENER_DOMAINS:
    SHORTENER_DOMAINS_TRIE.add(domain)


def is_l_shortened_domain(url):
    parsed = safe_urlsplit(url)

    return parsed.hostname.startswith("l.") and re.search(
        DOMAIN_STARTS_L_RE, parsed.path
    )


def is_shortened_url(url):
    parsed = safe_urlsplit(url)

    # NOTE: shortener domain homepages are not shortened urls per se
    if is_homepage(parsed):
        return False

    # Shortener domains starting with 'l.'
    if is_l_shortened_domain(parsed):
        return True

    return SHORTENER_DOMAINS_TRIE.match(parsed)



# -*- coding: utf-8 -*-
# =============================================================================
# Ural Should Resolve Function
# =============================================================================
#
# Function testing whether the given url looks like something we should resolve.
#
from ural.classes.hostname_trie_set import HostnameTrieSet
from ural.is_shortened_url import SHORTENER_DOMAINS, is_l_shortened_domain
from ural.is_homepage import is_homepage
from ural.utils import safe_urlsplit

SHOULD_RESOLVE_DOMAINS = ["doi.org", "list-manage.com"]
SHOULD_RESOLVE_TRIE = HostnameTrieSet()

for domain in SHORTENER_DOMAINS + SHOULD_RESOLVE_DOMAINS:
    SHOULD_RESOLVE_TRIE.add(domain)


def should_resolve(url):
    parsed = safe_urlsplit(url)

    # NOTE: shortener domain homepages are not shortened urls per se
    if is_homepage(parsed):
        return False

    # Shortener domains starting with 'l.'
    if is_l_shortened_domain(parsed):
        return True

    return SHOULD_RESOLVE_TRIE.match(parsed)



# =============================================================================
# Ural URL Extraction From HTML Function
# =============================================================================
#
# A function returning an iterator over the urls present in the HTML string
# argument.
#
from __future__ import unicode_literals
import re
from ural.patterns import URL_IN_HTML_RE
from ural.utils import urljoin


def clean_link(link):
    """Removes leading and trailing whitespace and punctuation"""
    return link.strip("\t\r\n '\"\x0c")


def urls_from_html(string, base_url=None):
    """
    Function returning an iterator over the urls present in the HTML string argument.

    Args:
        string (str): source html string.
        base_url (str, optional): base_url to concatenate to the found urls.
            Defaults to None.

    Yields:
        str: an url.

    """
    for match in re.finditer(URL_IN_HTML_RE, string):
        url = clean_link(match.group(1))

        if base_url is not None:
            url = urljoin(base_url, url)

        yield url



# =============================================================================
# Ural URL Normalization Function
# =============================================================================
#
# A handy function relying on heuristics to find and drop irrelevant or
# non-discriminant parts of a URL.
#
import re
from os.path import splitext

from ural.data import ISO_3166_1_COUNTRIES_ALPHA_2
from ural.ensure_protocol import ensure_protocol
from ural.infer_redirection import infer_redirection as resolve
from ural.utils import (
    parse_qsl,
    quote,
    urlsplit,
    urlunsplit,
    decode_punycode_hostname,
    unquote,
    normpath,
    fix_common_query_mistakes,
    SplitResult,
)
from ural.patterns import PROTOCOL_RE

RESERVED_CHARACTERS = ";,/?:@&=+$"
UNRESERVED_CHARACTERS = "-_.!~*'()"
SAFE_CHARACTERS = RESERVED_CHARACTERS + UNRESERVED_CHARACTERS

IRRELEVANT_QUERY_PATTERN = r"^(?:__twitter_impression|_guc_consent_skip|guccounter|fb_action_types|fb_action_ids|fb_source|echobox|feature|recruiter|_unique_id|twclid|mibextid|campaignid|adgroupid|cn-reloaded|ao_noptimize|mkt_tok|fbclid|igshid|refid|gclid|mc_cid|mc_eid|__tn__|_ft_|dclid|wpamp|fref|usqp|ncid|mtm_.+|utm_.+%s|s?een|xt(?:loc|ref|cr|np|or|s)|at_.+|_ga)$"

IRRELEVANT_SUBDOMAIN_PATTERN = r"\b(?:www\d?|mobile%s|m)\."

AMP_QUERY_PATTERN = r"|amp_.+|amp"
AMP_QUERY_COMBOS = {"outputtype": ("amp",)}
AMP_SUBDOMAIN_PATTERN = r"|amp"
AMP_SUFFIXES_RE = re.compile(r"(?:\.amp(?=\.html$)|\.amp/?$|(?<=/)amp/?$)", re.I)

IRRELEVANT_QUERY_RE = re.compile(IRRELEVANT_QUERY_PATTERN % r"", re.I)
IRRELEVANT_SUBDOMAIN_RE = re.compile(IRRELEVANT_SUBDOMAIN_PATTERN % r"", re.I)

IRRELEVANT_QUERY_AMP_RE = re.compile(IRRELEVANT_QUERY_PATTERN % AMP_QUERY_PATTERN, re.I)
IRRELEVANT_SUBDOMAIN_AMP_RE = re.compile(
    IRRELEVANT_SUBDOMAIN_PATTERN % AMP_SUBDOMAIN_PATTERN, re.I
)

IRRELEVANT_QUERY_COMBOS = {
    "marfeeltn": ("amp",),
    "mode": ("amp",),
    "output": ("amp",),
    "platform": ("hootsuite",),
    "fromref": ("twitter",),
    "m": (
        "0",
        "1",
    ),
    "ref": set(
        [
            "bookmark",
            "bookmarks",
            "distributor_share",
            "fb",
            "fb_i",
            "m_notif",
            "nf",
            "notif",
            "shortener",
            "ts",
            "tw",
            "tw_i",
            "twhr",
            "twhs",
            "twitter",
            "viral",
            "feed",
            "twtrec",
        ]
    ),
    "s": lambda v: len(v) <= 2 and all("0" <= x <= "9" for x in v),
    "source": ("twitter",),
    "sns": ("tw",),
    "spref": ("fb", "ts", "tw", "tw_i", "twitter"),
    "_ss": ("r",),
}

# NOTE: if this list becomes too long, switch to HostnameTrieMap
PER_DOMAIN_QUERY_FILTERS = [
    ("facebook.com", lambda k, v: k == "_rdc" or k == "_rdr"),
]

LANG_QUERY_KEYS = ("gl", "hl")


def stringify_qs(item):
    if item[1] == "":
        return item[0]

    return "%s=%s" % item


def should_strip_query_item(
    item, normalize_amp=True, strip_lang_query_items=False, domain_filter=None
):
    key = item[0].lower()

    pattern = IRRELEVANT_QUERY_AMP_RE if normalize_amp else IRRELEVANT_QUERY_RE

    if pattern.match(key):
        return True

    value = item[1]

    if key in IRRELEVANT_QUERY_COMBOS:
        result = IRRELEVANT_QUERY_COMBOS[key]
        if callable(result):
            return result(value)
        return value in IRRELEVANT_QUERY_COMBOS[key]

    # NOTE: only keep elif because query combos and amp query combos
    # are mutually exclusive.
    elif normalize_amp and key in AMP_QUERY_COMBOS:
        return value in AMP_QUERY_COMBOS[key]

    if strip_lang_query_items and key in LANG_QUERY_KEYS:
        return True

    if domain_filter is not None:
        return domain_filter(key, value)

    return False


def should_strip_fragment(fragment):
    if fragment == "!/" or fragment == "/" or fragment == "!":
        return False

    return fragment.startswith("/") or fragment.startswith("!")


def strip_lang_subdomains_from_netloc(netloc):
    if netloc.count(".") > 1:
        subdomain, remaining_netloc = netloc.split(".", 1)
        if len(subdomain) == 5 and "-" in subdomain:
            lang, country = subdomain.split("-", 1)
            if len(lang) == 2 and len(country) == 2:
                if (
                    lang.upper() in ISO_3166_1_COUNTRIES_ALPHA_2
                    and country.upper() in ISO_3166_1_COUNTRIES_ALPHA_2
                ):
                    netloc = remaining_netloc
        elif len(subdomain) == 2:
            if subdomain.upper() in ISO_3166_1_COUNTRIES_ALPHA_2:
                netloc = remaining_netloc

    return netloc


def normalize_url(
    url,
    unsplit=True,
    sort_query=True,
    strip_authentication=True,
    strip_trailing_slash=True,
    strip_index=True,
    strip_protocol=True,
    strip_irrelevant_subdomains=True,
    strip_lang_subdomains=False,
    strip_lang_query_items=False,
    strip_fragment="except-routing",
    normalize_amp=True,
    fix_common_mistakes=True,
    infer_redirection=True,
    quoted=True,
):
    """
    Function normalizing the given url by stripping it of usually
    non-discriminant parts such as irrelevant query items or sub-domains etc.

    This is a very useful utility when attempting to match similar urls
    written slightly differently when shared on social media etc.

    Args:
        url (str): Target URL as a string.
        sort_query (bool, optional): Whether to sort query items or not.
            Defaults to `True`.
        strip_authentication (bool, optional): Whether to drop authentication.
            Defaults to `True`.
        strip_trailing_slash (bool, optional): Whether to drop trailing slash.
            Defaults to `False`.
        strip_index (bool, optional): Whether to drop trailing index at the end
            of the url. Defaults to `True`.
        strip_irrelevant_subdomains (bool, optional): Whether to strip irrelevant subdomains such as www etc.
            Default to True.
        strip_lang_subdomains (bool, optional): Whether to drop language subdomains
            (ex: 'fr-FR.lemonde.fr' to only 'lemonde.fr' because 'fr-FR' isn't a relevant subdomain, it indicates the language and the country).
            Defaults to `False`.
        strip_protocol (bool, optional): Whether to strip the url's protocol.
            Defaults to `True`.
        strip_fragment (bool|str, optional): Whether to drop non-routing fragment from the url?
            If set to `except-routing` will only drop non-routing fragment (i.e. fragments that
            do not contain a "/").
            Defaults to `except-routing`.
        normalize_amp (bool, optional): Whether to attempt to normalize Google
            AMP urls. Defaults to True.
        fix_common_mistakes (bool, optional): Whether to attempt solving common mistakes.
            Defaults to True.
        infer_redirection (bool, optional): Whether to attempt resolving common
            redirects by leveraging well-known GET parameters. Defaults to `False`.
        quoted (bool, optional): Normalizing to quoted or unquoted.
            Defaults to True.

    Returns:
        string: The normalized url.

    """
    original_url_arg = url

    if infer_redirection:
        url = resolve(url)

    if isinstance(url, SplitResult):
        has_protocol = bool(url.scheme)
        splitted = url
    else:
        url = url.strip()
        has_protocol = PROTOCOL_RE.match(url)

        # Ensuring scheme so parsing works correctly
        if not has_protocol:
            url = "http://" + url

        # Parsing
        try:
            splitted = urlsplit(url)
        except ValueError:
            return original_url_arg

    scheme, netloc, path, query, fragment = splitted

    # Fixing common mistakes
    if fix_common_mistakes and query:
        query = fix_common_query_mistakes(query)

    # Handling punycode
    netloc = decode_punycode_hostname(netloc)

    # Dropping :80 & :443
    if netloc.endswith(":80"):
        netloc = netloc[:-3]
    elif netloc.endswith(":443"):
        netloc = netloc[:-4]

    # Normalizing the path
    if path:
        trailing_slash = False
        if path.endswith("/") and len(path) > 1:
            trailing_slash = True
        path = normpath(path)
        if trailing_slash and not strip_trailing_slash:
            path = path + "/"

    # Handling Google AMP suffixes
    if normalize_amp:
        path = AMP_SUFFIXES_RE.sub("", path)

    # Dropping index:
    if strip_index:
        segments = path.rsplit("/", 1)

        if len(segments) != 0:
            last_segment = segments[-1]
            filename, ext = splitext(last_segment)

            if filename == "index":
                segments.pop()
                path = "/".join(segments)

    # Dropping irrelevant query items
    if query:
        domain_filter = None

        if splitted.hostname:
            domain_filter = next(
                (
                    f
                    for d, f in PER_DOMAIN_QUERY_FILTERS
                    if splitted.hostname.endswith(d)
                ),
                None,
            )

        qsl = parse_qsl(query, keep_blank_values=True)
        qsl = [
            stringify_qs(item)
            for item in qsl
            if not should_strip_query_item(
                item,
                normalize_amp=normalize_amp,
                strip_lang_query_items=strip_lang_query_items,
                domain_filter=domain_filter,
            )
        ]

        if sort_query:
            qsl = sorted(qsl)

        query = "&".join(qsl)

    # Dropping fragment if it's not routing
    if fragment and strip_fragment:
        if strip_fragment is True or not should_strip_fragment(fragment):
            fragment = ""

    # Always dropping trailing slash with empty query & fragment
    if path == "/" and not fragment and not query:
        path = ""

    # Dropping irrelevant subdomains
    if strip_irrelevant_subdomains:
        netloc = re.sub(
            IRRELEVANT_SUBDOMAIN_AMP_RE if normalize_amp else IRRELEVANT_SUBDOMAIN_RE,
            "",
            netloc,
        )

    # Dropping language as subdomains
    if strip_lang_subdomains:
        netloc = strip_lang_subdomains_from_netloc(netloc)

    # Dropping scheme
    if strip_protocol or not has_protocol:
        scheme = ""

    # Dropping authentication
    if strip_authentication:
        netloc = netloc.split("@", 1)[-1]

    # Normalizing AMP subdomains
    if normalize_amp and netloc.startswith("amp-"):
        netloc = netloc[4:]

    # Dropping trailing slash
    if strip_trailing_slash and path.endswith("/"):
        path = path.rstrip("/")

    # Quoting or not
    if quoted:
        path = quote(path)
        query = quote(query, RESERVED_CHARACTERS)
        fragment = quote(fragment, SAFE_CHARACTERS)
    else:
        path = unquote(path)
        query = unquote(query)
        fragment = unquote(fragment)

    # Result
    result = SplitResult(scheme, netloc.lower(), path, query, fragment)

    if not unsplit:
        return result

    # TODO: check if works with `unsplit=False`
    if strip_protocol or not has_protocol:
        result = urlunsplit(result)[2:]
    else:
        result = urlunsplit(result)

    return result


def normalize_hostname(hostname, normalize_amp=True, strip_lang_subdomains=False):
    hostname = hostname.strip().lower()

    pattern = IRRELEVANT_SUBDOMAIN_AMP_RE if normalize_amp else IRRELEVANT_SUBDOMAIN_RE

    hostname = pattern.sub("", hostname)

    if normalize_amp and hostname.startswith("amp-"):
        hostname = hostname[4:]

    hostname = decode_punycode_hostname(hostname)

    if strip_lang_subdomains:
        hostname = strip_lang_subdomains_from_netloc(hostname)

    return hostname


def get_normalized_hostname(
    url, normalize_amp=True, strip_lang_subdomains=False, infer_redirection=True
):
    if infer_redirection:
        url = resolve(url)

    if isinstance(url, SplitResult):
        splitted = url
    else:
        try:
            splitted = urlsplit(ensure_protocol(url.strip()))
        except ValueError:
            return None

    if not splitted.hostname:
        return None

    return normalize_hostname(
        splitted.hostname,
        normalize_amp=normalize_amp,
        strip_lang_subdomains=strip_lang_subdomains,
    )
